/*
*    cwtmp: compact a wtmp file by discarding LOGIN and GETTY entries
*    Discards entries of ut_type LOGIN_PROCESS and INIT_PROCESS,
*    and entries with ut_name that begins with LOGIN.
*    90/11/16 John H. DuBois III
*    93/02/28 added entry discarding based on name prefix
*    94/03/14 Modified error message to reflect ambiguity of getutent() failure
*    94/06/22 Exit on error.  Expanded help.
*    94/07/03 Process command line options.  Exit 0 at end of main().
*    94/07/03 Warn about bad size of file, since this is a non-obvious thing
*             that makes getutent() fail.
*    94/10/23 Fixed so that /etc/wtmp really is the default.
*    94/12/14 Do not use utmp lib routines; poor error reporting & they
*             refuse to operate on a corrupt (wrong size) file.
*             Added all options.
*    95/07/23 Added prsw options.
*             Fixed bug that caused coredump if a ut_type < 0 was returned.
*    95/10/15 Added it options.
*    95/12/30 Write help to stdout, not stderr.
*    96/01/06 Began adding e option.
*/

#include <time.h>
#include <stdio.h>
#include <pwd.h>
#include <utmp.h>
#ifdef EXTENDED
#   include <utmpx.h>
#endif
#include "nutmp.h"

char **StrToArr(char *S,char Sep[],char EscChar);

char SCCSstring[] = "@(#) cwtmp 2.2 95/10/15";

main(argc,argv)
int argc;
char *argv[];
{
    struct utmp *Ent;
    char *InFile = WTMP_FILE;
    char *DeleteName = NULL;
    int DeleteLen = 0;
    static int RemoveTypes[UTMAXTYPE+1];
    int TypesGiven = 0;
    long Skip = 0;
    long Write = 0;
    long Value;
    int Readable = 0;
    char **IDlist = NULL;
    char **id;
    int DelInitCmds = 0;
    int Extended = 0;

    int Err = 0;
    char c;			/* Option character */
    extern char *optarg;	/* For optargs */
    extern int optind;		/* For optargs */

/*"-e: Process an extended utmp file (e.g. utmpx, wtmpx).\n"\ */

    while ((c = getopt(argc,argv,"ad:efhi:nrp:w:s:tvx0123456789")) != -1)
	switch (c) {
	case 'h':
	    fputs(
"cwtmp: compress wtmp file data by removing uninteresting entries.\n"\
"Usage: cwtmp [-ahnrvx] [-d<prefix>] -i<id[:id...]> [-[psw]<numrec>] [file]\n"\
"cwtmp compresses a wtmp file by discarding the uninteresting utmp entries\n"\
"that make up a large fraction of the typical wtmp file.  If no filename is\n"\
"given, /etc/wtmp is read.  utmp entries in the file are written to the\n"\
"output, except entries of type LOGIN_PROCESS and INIT_PROCESS, which are\n"\
"skipped.\n"\
"If the input file is not a multiple of the size of a utmp record and one\n"\
"of the v, a, or f flags is not given, cwtmp will exit with an error\n"\
"message and nonzero status.\n"\
"Options:\n"\
"-i<id[:id...]>: Discard entries whose id field is one of those given in\n"\
"    the colon-separated list.\n"\
"-t: Discard entries which have a non-empty user field and an empty line\n"\
"    field.  These are usually commands run from inittab which are not\n"\
"    associated with a tty.\n"\
"-v: Check every utmp record for validity.  If an invalid record is found,\n"\
"    scan the input character by character attempting to find the start of\n"\
"    the next utmp record.  Invalid records are not output.\n"\
"-a: Autoverify.  If the input file has a size that is not a multiple of\n"\
"    the size of a utmp record, the -v option is turned on.\n"\
"-x: Turn on verbose verification operation.  Various diagnostics are\n"\
"    printed, including a description of which verification tests fail.\n"\
"-d<name-prefix>: Entries whose ut_user field begins with the prefix are\n"\
"    discarded.\n"\
"-f: Do not fail if the input file does not have a valid size, and do no\n"\
"    checking for utmp record validity.  This is liable to produce unuseful\n"\
"    output if the input becomes unsynchronized, since the type test will\n"\
"    likely be done on a field component that is not the type.\n"\
"-n: Do not discard LOGIN_PROCESS and INIT_PROCESS entries.  This can be\n"\
"    used to fix a file with the -v or -a option without compressing it.\n"\
"-p<num-records>: Print a progress report to stderr after reading each\n"\
"    <num-records> records.\n"\
"-s<num-records>: Skip the initial <num-records> records in the file\n"\
"    without looking at them.\n"\
"-w<num-records>: Write <num-records> and then quit.\n"\
"-r: Print records in readable form, rather that writing the utmp structs.\n"\
"-0..-9: Remove entries with the given numeric type.  If any of these\n"\
"    options are given, entries of type LOGIN_PROCESS and INIT_PROCESS are\n"\
"    not removed unless their flags are explicitely given.  See utmp(F) for\n"\
"    a description of the types.\n",stdout);
	    exit(0);
	case 'd':
	    DeleteName = optarg;
	    DeleteLen = strlen(DeleteName);
	    break;
#ifdef EXTENDED
	case 'e':
	    utmpConfig(UTMP_EXTENDED,NULL);
	    Extended = 1;
	    break;   
#endif
	case 'i':
	    if (!(IDlist = StrToArr(optarg,":",'\0'))) {
		fprintf(stderr,"cwtmp: bad id string.\n");
		exit(1);
	    }
	    break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	    RemoveTypes[c - '0'] = 1;
	    TypesGiven = 1;
	    break;
	case 'n':
	    TypesGiven = 1;
	    break;
	case 'r':
	    Readable = 1;
	    break;
	case 'a':
	    utmpConfig(UTMP_AUTOVERIFY,NULL);
	    break;
	case 'f':
	    utmpConfig(UTMP_NOSIZECHECK,NULL);
	    break;
	case 's': case 'p': case 'w':
	    if ((Value = atol(optarg)) < 1) {
		fprintf(stderr,
"cwtmp: argument to -%c must be a positive integer.  Use -h for help.\n",c);
		exit(1);
	    }
	    switch (c) {
	    case 's':
		Skip = Value;
		break;
	    case 'p':
		utmpConfig(UTMP_PROGRESS,&Value);
		break;
	    case 'w':
		Write = Value;
		break;
	    }
	    break;
	case 't':
	    DelInitCmds = 1;
	    break;
	case 'x':
	    utmpConfig(UTMP_VERBOSE,NULL);
	    break;
	case 'v':
	    utmpConfig(UTMP_VERIFY,NULL);
	    break;
	case '?':
	    Err = 1;
	}
    if (Err)
	exit(1);
    switch (argc - optind) {
    case 0:
	InFile = Extended ? "/etc/wtmpx" : "/etc/wtmp";
	break;
    case 1:
	InFile = argv[optind];
	break;
    default:
	fputs("cwtmp: too many arguments.  Use -h for help.\n",stderr);
	exit(1);
    }
    n_utmpname(InFile);
    if (!(Ent = n_getutent(Skip)))
	exit(1);
    if (!TypesGiven) {
	RemoveTypes[LOGIN_PROCESS] = 1;
	RemoveTypes[INIT_PROCESS] = 1;
    }
    /*	    strncmp(Ent->ut_user,"LOGIN",5) && */
    do
	/* Remove record if its type is invalid or is in the list of types
	   to remove, or the user name begins with the prefix that was given
	   with -d, or its id was given with -i, or if -t was given and
	   the entry has a non-empty user field and an empty line field */
	if (Ent->ut_type >= 0 && Ent->ut_type <= UTMAXTYPE &&
	!RemoveTypes[Ent->ut_type] &&
	(!DeleteLen || strncmp(Ent->ut_user,DeleteName,DeleteLen)) &&
	(!DelInitCmds || !*Ent->ut_user || *Ent->ut_line)) {
	    if (IDlist) {
		for (id = IDlist; *id && strncmp(*id,Ent->ut_id,4); id++)
		    ;
		if (*id)
		    continue;
	    }
	    if (Readable)
		utmpPrint(Ent,stdout);
	    else if (write(1,Ent,sizeof(struct utmp)) == -1) {
		perror(*argv);
		exit(1);
	    }
	    if (Write && !--Write)
		exit(0);
	}
    while (Ent = n_getutent(0));
    exit(0);
}
