/*
    nutmp.c: replacement for standard library utmp routines, which have poor
    error reporting and refuse to operate on a corrupt (wrong size) file.
    94/12/12 John H. DuBois III (john@armory.com)
    95/07/23 Added Progress and skip
    96/04/13 Require ut_time to be < current_time+7200
*/

#include <stdio.h>
#include <sys/types.h>
#include <utmp.h>
#include <time.h>
#ifdef EXTENDED
#   include <utmpx.h>
#endif
#include <sys/stat.h>
#include <string.h>
#include <sys/param.h>
#include <ctype.h>
#include "nutmp.h"

#define UBUFSIZE 10

char *_utmpName = "/etc/utmp";
FILE *_utmpFile = NULL;
int Verify,AutoVerify,Verbose;
int SizeFail = 1;
struct utmp Buffer[UBUFSIZE];
struct utmp *u = Buffer;
long Offset, RecordNum;
long SkippedBytes;
long Progress = 0;
long ProgressCt;
int Extended = 0;
int utsize = sizeof(struct utmp);
char *utstructName = "utmp";
time_t TestTime = (time_t) 0;

/*
*    Change the name of the file read by the utmp routines.
*/
void n_utmpname(char *File) {
    _utmpName = strdup(File);
}

int utmpConfig(int a, void *Arg)
{
    switch (a) {
    case UTMP_VERBOSE:
	Verbose = 1;
	break;
    case UTMP_PROGRESS:
	Progress = ProgressCt = *((long *)Arg);
	break;
    case UTMP_VERIFY:
	Verify = 1;
	SizeFail = 0;
	break;
    case UTMP_NOSIZECHECK:
	SizeFail = 0;
	break;
#ifdef EXTENDED
    case UTMP_EXTENDED:
	Extended = 1;
	utsize = sizeof(struct utmpx);
	utstructName = "utmpx";
	break;
#endif
    case UTMP_AUTOVERIFY:
	AutoVerify = 1;
	SizeFail = 0;
	break;
    }
    return 0;	/* for now */
}

/*
*    Returns 1 on failure, 0 on success
*/
static int openUtmp()
{
    struct stat buf;

    if (!stat(_utmpName, &buf) && buf.st_size % utsize) {
	fprintf(stderr,
	"%s: size of %s (%ld bytes) is not a multiple of the size of a\n"\
	"%s record (%d bytes).\n",SizeFail ? "Error" : "Warning",
	_utmpName,buf.st_size,utstructName,utsize);
	if (SizeFail)
	    return 1;
	if (AutoVerify)
	    Verify = 1;
    }
    if ((_utmpFile = fopen(_utmpName,"r")) == NULL) {
	fprintf(stderr,"Could not open file \"%s\" for reading: \n",_utmpName);
	fflush(stderr);
	perror("");
	return 1;
    }
    if (Verbose)
	fprintf(stderr,"Successfully opened %s file %s\n",utstructName,
	_utmpName);
    Offset = RecordNum = 0;
    return 0;
}

static char *NullTerm(char *s, int n)
{
    static char s2[13];
    
    strncpy(s2,s,n);
    s2[n] = '\0';
    return s2;
}

void utmpPrint(struct utmp *u, FILE *f)
{
    fprintf(f,"user=%-8s ",NullTerm(u->ut_user,8));
    fprintf(f,"id=%-4s ",NullTerm(u->ut_id,4));
    fprintf(f,"line=%-12s ",NullTerm(u->ut_line,12));
    fprintf(f,"pid=%-5d ",(int) u->ut_pid);
    fprintf(f,"type=%-1d ",(int) u->ut_type);
    fprintf(f,"time=%-10ld\n",(long) u->ut_time);
}

/*
*    Returns null if s is a string of 0 or more printable chars followed by
*    0 or more nulls, with the total string length being n
*    Otherwise, returns the invalid character encountered.
*/
static char sTest(char *s, int n)
{
    while (n-- && isprint(*(s++)))
	;
    if (n == -1)
	return '\0';	/* All printable chars */
    n++;
    s--;
    /* Test if all chars after printable chars are nulls */
    while (n-- && !*(s++))
	;
    return *(--s);
}

/*
*   Returns true if u is a good utmp structure, according to these tests:
*   utmp struct elements:
*   	char ut_user[8]		Printable chars followed by nulls
*   	char ut_id[4]		Printable chars followed by nulls
*   	char ut_line[12]	Printable chars followed by nulls
*   	short ut_pid		Integer 0..30000
*   	short ut_type		Integer 0..UTMAXTYPE
*   	struct exit_status
*   	    short e_termination	Integer
*   	    short e_exit	Integer
*   	time_t ut_time		Integer > 0, < approx-current-time+7200
*/
static int utmpTest(struct utmp *u)
{
    unsigned char c;

    if (!sTest(u->ut_user,8) && !sTest(u->ut_id,4) && !sTest(u->ut_line,12)
    && 0 <= u->ut_pid && u->ut_pid <= MAXPID && 0 <= u->ut_type &&
    u->ut_type <= UTMAXTYPE && u->ut_time > 0)
	return 1;
    else {
	if (Verbose) {
	    if (Progress)
		fputc('\n',stderr);
	    fputs("Bad: ",stderr);
	    if (c = sTest(u->ut_user,8))
		fprintf(stderr,"user (char 0x%02x) ",(unsigned) c);
	    if (c = sTest(u->ut_id,4))
		fprintf(stderr,"id (char 0x%02x) ",(unsigned) c);
	    if (c = sTest(u->ut_line,12))
		fprintf(stderr,"line (char 0x%02x) ",(unsigned) c);
	    if (!(0 <= u->ut_pid && u->ut_pid <= MAXPID))
		fputs("pid ",stderr);
	    if (!(0 <= u->ut_type && u->ut_type <= UTMAXTYPE))
		fputs("type ",stderr);
	    /* If time fails TestTime test, get time again in case it's just
	       because we've been processing a long time.  Actually, 
	       TestTime is initialized to 0 and only set here. */
	    if (!(u->ut_time > 0 && u->ut_time < TestTime &&
	    u->ut_time < (TestTime = time(NULL) + 7200) ))
		fputs("time ",stderr);
	    fputc('\n',stderr);
	    utmpPrint(u,stderr);
	}
	return 0;
    }
}

static struct utmp *utmpSearch()
{
    int c;
    char *p;
    long StartOffset = Offset;

    if (Progress)
	fputc('\n',stderr);
    fprintf(stderr,
    "%s read out of sync at record %ld (byte offset %lu); searching...\n",
    utstructName,RecordNum,Offset);
    p = (char *) u;	/* so we can do 1-byte increments */
    while ((c = fgetc(_utmpFile)) != EOF) {
	Offset++;
	/* if we've reached the end of the buffer, copy the current structure
	   back to the beginning and point p there */
	if (p == (char *) (Buffer + (UBUFSIZE - 1))) {
	    memcpy(Buffer,p,utsize);
	    p = (char *) Buffer;
	}
	*(p + utsize) = (char) c;
	p++;
	if (utmpTest((struct utmp *)p)) {
	    u = (struct utmp *)p;
	    fprintf(stderr,
	    "%s read back in sync; skipped %lu bytes.  Synced record:\n",
	    utstructName,Offset - StartOffset);
	    SkippedBytes += Offset - StartOffset;
	    utmpPrint(u,stderr);
	    return u;
	}
    }
    u = (struct utmp *)p;
    fprintf(stderr,
    "%s read failed to sync; at EOF after skipping %ul bytes.\n",
    utstructName,Offset - StartOffset);
    return NULL;
}

/* Skip NumSkip records before reading the next record.
*  Return value: pointer to the next utmp record on success;
*  0 on failure (including EOF).
*/
struct utmp *n_getutent(long NumSkip)
{
    struct utmp *Ret;

    if (!_utmpFile && openUtmp())
	return (struct utmp *)0;
    if (NumSkip)
	if (fseek(_utmpFile,NumSkip*utsize,SEEK_CUR)) {
	    if (Progress)
		fputc('\n',stderr);
	    fprintf(stderr,"Skip of %ld records in file \"%s\" failed.\n",
	    NumSkip,_utmpName);
	}
	else {
	    Offset += NumSkip * utsize;
	    RecordNum += NumSkip;
	    SkippedBytes += NumSkip * utsize;
	}
    if (fread(u,utsize,1,_utmpFile) == 1) {
	if (Verify && !utmpTest(u))
	    Ret = utmpSearch();
	else
	    Ret = u;
    }
    else
	Ret = (struct utmp *)0;
    Offset += utsize;
    RecordNum++;
    if (!Ret) {
	if (Progress)
	    fputc('\n',stderr);
	if (SkippedBytes)
	    fprintf(stderr,"Skipped %ul bytes total in file %s\n",SkippedBytes,
	    _utmpName);
	return (struct utmp *) 0;
    }
    if (Progress && !--ProgressCt) {
	fprintf(stderr,"%ld records\r",RecordNum);
	ProgressCt = Progress;
    }
    return Ret;
}
